# This R code tests all time series in the GPDD for cyclicity vs the OUSS null model as well as white noise.
# Execute using source("GPDD_main.R") in the console.
# GPDD should be in the form of separate files, one for each time series (specified below)
# 	Each time series file name should begin with "GPDD_"
#	Comments in time series should be preceeded by '#'
#	Time series should consist of two comma-separated columns (time and signal values).
#	No other uncommented text is allowed.
# You can mess with some of the parameters of the analysis below.
# The actual peak detection and statistical evaluation is done using the peacots package, available on CRAN.
#
# Stilianos Louca
# October 26, 2014


library(peacots)



#####################################
# Options

minDataSize 				= 25; 		# Ignore GPDD time series of smaller size
OUSSsignificanceLevel 		= 0.05;
WNsignificanceLevel	 		= 0.05;
GPDDdir 					= "GPDD";	# Directory with GPDD time series files
GPDDfilePrefix				= "GPDD_";	# Prefix of individual GPDD time series files
plotTimeSeries 				= TRUE;		# Whether to plot and save a detailed report of each analyzed time series.
log_transform_data 			= FALSE;	# Whether to log-transform time series data. If true, then time series with non-positive entries will be skipped.
detrend_WN 					= TRUE;		# Whether to detrend time series before testing against WN (not relevant for OUSS)

minPeakFreq_OUSS			= 0.0;		# Only search for periodogram peak among frequencies greater or equal to this threshold. Lower-frequency modes are omitted. Only used for OUSS test. Note that zero-mode is always discarded.
minFitFreq_OUSS				= 0.0;		# Fit OUSS & WN models using only frequencies greater or equal to this threshold. Only used for OUSS test. Note that zero-mode is always discarded.
minPeakFreq_WN				= 0.0;		# Similar to minPeakFreq_OUSS, but for WN test. Note that zero-mode is always discarded.
minFitFreq_WN				= 0.0;		# Similar to minFitFreq_OUSS, but for WN test. Note that zero-mode is always discarded.

skipSingleSidesPeaks_WN		= TRUE; 	# If true, then cases where the periodogram peak is not trully a peak (i.e. the next-lowest frequency does not have lower power), are skipped
skipSingleSidesPeaks_OUSS	= TRUE; 	# 







###########################################
#### Auxiliary functions



getNonExistingFile = function(parentDir, baseName){
	DEFAULT_NUMWIDTH = 2;
	n = 0;
	while(TRUE){
		n = n+1;
		cfile = paste0(parentDir, "/", baseName, formatC(n, width=DEFAULT_NUMWIDTH, flag="0"));
		if(!file.exists(cfile)){ return(cfile); }
	}
}



linearExtrapolate = function(x1, x2, y1, y2, x){
	return(y1 + (y2-y1)*(x-x1)/(x2-x1));
}


# Calculate cumulative distribution function (CDF) from samples
# All Xvalues on the left of all samples will be assigned a CDF=0, all Xvalues to the right of all samples will be assigned a CDF=1.
# Returns a vector of same size as Xvalues[] with corresponding CDF values
getEmpiricalCumulativeDF = function(samples, 	 # (input) vector of samples (x-values) from which to estimate cumulative distribution function. Samples might get re-ordered.
									Xvalues){ 	 # (input) vector of X-values for which to estimate CDF. Must be in increasing order.
	NS = length(samples);
	NX = length(Xvalues);
	CDFvalues = rep(0,NX);
	samples = sort(samples);
	xi = 1;
	for(si in 1:NS){
		if(xi>NX) break;
		while(samples[si]>Xvalues[xi]){
			if(si==1){
				CDFvalues[xi] = 0;
			}else{
				CDFvalues[xi] = linearExtrapolate(samples[si-1], samples[si], (si-1)/NS, si/NS, Xvalues[xi]);
			}
			xi = xi+1;
			if(xi>NX) break;
		}
	}
	if(xi<=NX){
		for(i in xi:NX){
			CDFvalues[i] = 1;
		}
	}
	return(CDFvalues);
}







# Function for plotting peak analysis results for a particular time series
plotPeakAnalysis = function(name, times, signal_raw, signal_OUSS, signal_WN, peak_OUSS, peak_WN, dont_show_OUSS_fit, dont_show_WN_fit, plotDir){
	old.par <- par(mfrow=c(2, 3))
	delta = (tail(times,n=1)-times[1])/(length(times)-1); # time step
	
	
	#plot time series
	plot(ts(times), ts(signal_raw), xy.label=FALSE, type="l", ylab="Value", xlab="time", main=sprintf("Time series raw (%s)", name), cex=0.8, cex.main=0.9)
	

	# OUSS
	
	#plot time series OUSS
	plot(ts(times), ts(signal_OUSS), xy.label=FALSE, type="l", ylab="Value", xlab="time", main=sprintf("Time series\n(used for OUSS analysis)"), cex=0.8, cex.main=0.9)
			
	if(!peak_OUSS$error){	

		#plot periodogram OUSS
		plot(ts(peak_OUSS$frequencies), ts(peak_OUSS$periodogram), xy.label=FALSE, type="l", ylab="Power", xlab="frequency", main=ifelse(dont_show_OUSS_fit!="", sprintf("Power spectrum (%s)\n(%s)", name, dont_show_OUSS_fit), sprintf("Power spectrum (%s)\n(fmax=%.3g, POUSS=%.2g)", name,peak_OUSS$frequencies[peak_OUSS$peakMode],peak_OUSS$P)), col="black", cex=0.8, cex.main=0.9);
	
		if(dont_show_OUSS_fit==""){
			#plot fitted OUSS periodogram
			lines(peak_OUSS$frequencies[peak_OUSS$minFitMode:length(peak_OUSS$frequencies)], peak_OUSS$fittedPS[peak_OUSS$minFitMode:length(peak_OUSS$fittedPS)], col="red");
		
			#plot legend
			legend((0.6*peak_OUSS$frequencies[1]+0.4*tail(peak_OUSS$frequencies,1)), (0.85*max(peak_OUSS$periodogram)), c("periodogram", "fitted OUSS"), lty=c(1,1), col=c("black", "red"), bty="n", cex=0.8)
		}
	}
	
	# WN
	
	#plot time series WN
	plot(ts(times), ts(signal_WN), xy.label=FALSE, type="l", ylab="Value", xlab="time", main=sprintf("Time series\n(used for WN analysis)"), cex=0.8, cex.main=0.9)
	
	if(!peak_WN$error){	
		#plot periodogram WN
		plot(ts(peak_WN$frequencies), ts(peak_WN$periodogram), xy.label=FALSE, type="l", ylab="Power", xlab="frequency", main=ifelse(dont_show_WN_fit!="", sprintf("Power spectrum (%s)\n(%s)", name, dont_show_WN_fit), sprintf("Power spectrum (%s)\n(fmax=%.3g, PWN=%.2g)", name,peak_WN$frequencies[peak_WN$peakMode],peak_WN$P)), col="black", cex=0.8, cex.main=0.9);
		
		if(dont_show_WN_fit==""){
			#plot fitted white noise periodogram
			lines(c(peak_WN$frequencies[peak_WN$minFitMode],tail(peak_WN$frequencies,1)), c(peak_WN$powerEstimate, peak_WN$powerEstimate ), col="blue");
			
			#plot legend
			legend((0.6*peak_WN$frequencies[1]+0.4*tail(peak_WN$frequencies,1)), (0.85*max(peak_WN$periodogram)), c("periodogram", "fitted WN"), lty=c(1,1), col=c("black", "blue"), bty="n", cex=0.8)
		}
	}
	
	par(old.par)
	
	#save plot as PDF
	if(plotDir != FALSE){
		dir.create(plotDir, showWarnings=FALSE);
		dev.copy2pdf(file=paste(plotDir, paste0(filename,"_plot.pdf"), sep="/"));
		
		# write data to file
		dataFile = paste(plotDir, paste0(filename,"_data.txt"), sep="/");
		cat(sprintf("# OUSS and WN analysis of times series %s\n# times\tsignal_raw\tsignal_OUSS\tsignal_WN\n", name), file=dataFile, append=FALSE);
		write.table(data.frame(times, signal_raw, signal_OUSS, signal_WN), , file=dataFile, append=TRUE, row.names=FALSE, col.names=FALSE, quote=FALSE, sep="\t");
		cat(sprintf("\n\n# periodogram of time series used for OUSS test\n# frequency\tpower\n"), file=dataFile, append=TRUE);
		write.table(data.frame(peak_OUSS$frequencies, peak_OUSS$periodogram), , file=dataFile, append=TRUE, row.names=FALSE, col.names=FALSE, quote=FALSE, sep="\t");
		cat(sprintf("\n\n# periodogram of time series used for WN test\n# frequency\tpower\n"), file=dataFile, append=TRUE);
		write.table(data.frame(peak_WN$frequencies, peak_WN$periodogram), , file=dataFile, append=TRUE, row.names=FALSE, col.names=FALSE, quote=FALSE, sep="\t");

	}
}






#############################################
#### GPDD analysis


#Look for GPDD time series files
dataFiles = list.files(path=paste(getwd(),GPDDdir,sep="/"), pattern=sprintf("^%s", GPDDfilePrefix), all.files=FALSE, full.names=FALSE, recursive=FALSE, ignore.case=TRUE);


# create new output dir and save copy of this script for future reference
outputDir = getNonExistingFile(paste0(dirname(GPDDdir), "/GPDD_runs"), "run_");
cat(sprintf("Analyzing GPDD.. All output will be written to '%s'\n", outputDir));
dir.create(outputDir, showWarnings=FALSE, recursive=TRUE);
file.copy(from="GPDD_main.R", to=outputDir, overwrite=TRUE);
plotDir = paste0(outputDir,"/plots");
dir.create(plotDir, showWarnings=FALSE, recursive=TRUE);
logFile = paste0(outputDir,"/log.txt");
scatterFile = paste0(outputDir,"/scatter_data.txt");


#Read GPDD time series from individual files and test each one for cyclicity
tooLowFreq_OUSS=0; errors_OUSS=0; errors_WN=0; cOUSS=cWN=0; failedLogTransform=0; tooSmall=0; peakAtFirstMode_OUSS=0; peakOnlySingleSided_WN=0; peakOnlySingleSided_OUSS=0;
FAPs_OUSS=c(); FAPs_WN = c();
count = 0; totalCount = length(dataFiles); feedbackPeriod = max(1, floor(totalCount/20));
dont_show_OUSS_fit=""; 
dont_show_WN_fit="";
OUSS_WN_RESULTS_TABLE = NULL;
for(filename in dataFiles){
	if(((count %% feedbackPeriod) == 0) && (count>0)){
		cat(sprintf("\n  %.2g%% finished..\n", (100.0 * count)/totalCount));
	}
	count = count+1;
	counted_this_OUSS = counted_this_WN = FALSE;
	
	#Load particular time series from data file
	timeSeries = read.csv(paste(GPDDdir,filename,sep="/"), header=FALSE, sep=",", comment.char="#");
	times = timeSeries[,1]; signal_raw = timeSeries[,2];
	time_step = (tail(timeSeries[,1],n=1)-timeSeries[1,1])/(length(timeSeries[,1])-1);
	
	#ignore time series if too small
	if(length(times)<minDataSize){ 
		tooSmall = tooSmall + 1; 
		next; 
	}
	
	# log transform if required
	if(log_transform_data){
		if(any(signal_raw<=1e-20)){
			 # skip this time series because of zero-entries
			failedLogTransform = failedLogTransform + 1;
			next;
		}
		signal = log(signal_raw);
	}else{
		signal = signal_raw;
	}
	
	#Find spectral peak and its statistical significance under the null hypothesis of an OUSS process
	peak_OUSS = peacots::evaluate.pm(times, signal, minPeakFreq=minPeakFreq_OUSS, minFitFreq=minFitFreq_OUSS);
	if(peak_OUSS$error){ 
		cat(sprintf("\nError in OUSS analysis of time series %s: %s\n\n", filename, peak_OUSS$errorMessage)); 
		errors_OUSS = errors_OUSS + 1;
		dont_show_OUSS_fit = paste0("error: ",peak_OUSS$errorMessage);
		counted_this_OUSS 	= FALSE;
	}else{
		#check statistical significance
		FAPs_OUSS 			= c(FAPs_OUSS, peak_OUSS$P);
		counted_this_OUSS 	= TRUE;
		dont_show_OUSS_fit 	= "";
		if(peak_OUSS$peakMode==1) peakAtFirstMode_OUSS = peakAtFirstMode_OUSS + 1;
		temp_peakOnlySingleSided_OUSS = (peak_OUSS$peakMode==1 || (peak_OUSS$periodogram[peak_OUSS$peakMode-1]>=peak_OUSS$periodogram[peak_OUSS$peakMode]) || (peak_OUSS$peakMode==length(peak_OUSS$frequencies)));
		if(temp_peakOnlySingleSided_OUSS){
			peakOnlySingleSided_OUSS = peakOnlySingleSided_OUSS + 1;
		}
		if((peak_OUSS$P<OUSSsignificanceLevel) && ((!temp_peakOnlySingleSided_OUSS) || (!skipSingleSidesPeaks_OUSS))){ 
			cOUSS = cOUSS + 1; 
		}
	}
	
	
	#Find spectral peak and its statistical significance under the null hypothesis of an OUSS process
	if(detrend_WN){
		fit=loess(Y~X,data.frame(X=times,Y=signal),span=0.5,degree=1);
		signal_WN = fit$residuals;
	}else{
		signal_WN = signal;
	}
	peak_WN = peacots::evaluate.pm.wn(times, signal_WN, minPeakFreq=minPeakFreq_WN, minFitFreq=minFitFreq_WN);
	if(peak_WN$error){ 
		cat(sprintf("\nError in WN analysis of time series %s: %s\n\n", filename, peak_WN$errorMessage)); 
		errors_WN = errors_WN + 1;
		dont_show_WN_fit = paste0("error: ",peak_WN$errorMessage);
		counted_this_WN 	= FALSE;
	}else{
		FAPs_WN 			= c(FAPs_WN, peak_WN$P); 
		counted_this_WN 	= TRUE;
		dont_show_WN_fit 	= "";
		temp_peakOnlySingleSided_WN = (peak_WN$peakMode==1 || (peak_WN$periodogram[peak_WN$peakMode-1]>=peak_WN$periodogram[peak_WN$peakMode]) || (peak_WN$peakMode==length(peak_WN$frequencies)));
		if(temp_peakOnlySingleSided_WN) peakOnlySingleSided_WN = peakOnlySingleSided_WN + 1;
		if((peak_WN$P<WNsignificanceLevel) && ((!temp_peakOnlySingleSided_WN) || (!skipSingleSidesPeaks_WN))){ 
			cWN = cWN + 1; 
		}
	}
	
	if(counted_this_OUSS && counted_this_WN){
		record = list(	ID=filename, 
						OUSS_FAP=peak_OUSS$P,
						WN_FAP=peak_WN$P, 
						OUSS_PM=peak_OUSS$peakMode, 
						WN_PM=peak_WN$peakMode, 
						OUSS_PeakModeAtMinMode=(peak_OUSS$peakMode==peak_OUSS$minPeakMode), 
						OUSS_PeakModeBelow4=(which.max(peak_OUSS$periodogram)<4),
						WN_PeakModeAtMinMode=(peak_WN$peakMode==peak_WN$minPeakMode),
						WN_PeakModeBelow4=(which.max(peak_WN$periodogram)<4),
						timeSeriesLength=length(times),
						time_step=time_step,
						OUSS_power_o=peak_OUSS$power_o,
						OUSS_lambda=peak_OUSS$lambda,
						OUSS_power_e=peak_OUSS$power_e,
						WN_powerEstimate=peak_WN$powerEstimate,
						peakOnlySingleSided_OUSS=temp_peakOnlySingleSided_OUSS,
						peakOnlySingleSided_WN=temp_peakOnlySingleSided_WN);
		if(is.null(OUSS_WN_RESULTS_TABLE)){ OUSS_WN_RESULTS_TABLE=data.frame(record, stringsAsFactors=FALSE); }
		else{ OUSS_WN_RESULTS_TABLE = rbind(OUSS_WN_RESULTS_TABLE, record); }
	}
	
	#plot individual time series & periodograms during calculations
	if(plotTimeSeries){
		plotPeakAnalysis(filename, times, signal_raw, signal, signal_WN, peak_OUSS, peak_WN, dont_show_OUSS_fit, dont_show_WN_fit, plotDir);
	}	
	
	cat(".");
}



#########################################################
# Evaluate analysis, do some statistics, plot some graphs



#Print summary statistics of GPDD cyclicity tests
result = sprintf(	"\n# Found %d GPDD time series. %d of them were too short (<%d), \
  %d of them could not be log-transformed. \
  Min considered peak frequency was %.3g for OUSS, %.3g for WN.\n\
  # Min considered fit frequency was %.3g for OUSS, %.3g for WN\n\
  # For OUSS analysis, %d could not be analyzed. Of the remaining %d time series, %d had their peak at minPeakMode, %d peaks were only one-sided %s \
  %d had their peak among the first 3 modes, %d (%.4g %%) were classified as cyclic.\n\
  # For WN analysis, %d could not be analyzed. Of the remaining %d time series, \
  %d had their peak at minPeakMode, %d peaks were only one-sided %s, \
  %d has their peak among first 3 modes, %d (%.4g %%) were classified as cyclic.",
					count, tooSmall, minDataSize, failedLogTransform, minPeakFreq_OUSS, minPeakFreq_WN, minFitFreq_OUSS, minFitFreq_WN, errors_OUSS,
					length(FAPs_OUSS), sum(OUSS_WN_RESULTS_TABLE[["OUSS_PeakModeAtMinMode"]]), peakOnlySingleSided_OUSS, ifelse(skipSingleSidesPeaks_OUSS," (and were skipped)", ""), sum(OUSS_WN_RESULTS_TABLE[["OUSS_PeakModeBelow4"]]), cOUSS, 100*cOUSS/length(FAPs_OUSS),
					errors_WN, length(FAPs_WN), sum(OUSS_WN_RESULTS_TABLE[["WN_PeakModeAtMinMode"]]), peakOnlySingleSided_WN, ifelse(skipSingleSidesPeaks_WN," (and were skipped)",""), sum(OUSS_WN_RESULTS_TABLE[["WN_PeakModeBelow4"]]), cWN, 100*cWN/length(FAPs_WN));
					
cat(sprintf("# %s\n%s\n", format(Sys.time(), "\n# %a %b %d, %Y, %X:"),result), file=logFile, append=TRUE);
cat(paste0(result,"\n"));



# Analyze CDF of estimated FAPs
old.par <- par(mfrow=c(1, 2))
Xvalues	= c(0.005,0.006,0.007,0.008,0.009,0.01,0.015,0.02,0.025,0.03,0.035,0.04,0.045,0.05,0.055,0.06,0.065,0.07,0.075,0.08,0.085,0.09,0.095,0.1,0.15,0.2,0.25,0.3,0.35,0.4,0.45,0.5,0.6,0.7,0.8,0.9,1);
CDFvalues_OUSS 	= getEmpiricalCumulativeDF(FAPs_OUSS, Xvalues);
CDFvalues_WN 	= getEmpiricalCumulativeDF(FAPs_WN, Xvalues);



# write log file
cat("\n\n# OUSS CDF of estimated FAPs:\n# FAP  CDF\n", file=logFile, append=TRUE);
write.table(data.frame(FAP=Xvalues,CDF=CDFvalues_OUSS), file=logFile, append=TRUE, row.names=FALSE, col.names=FALSE, quote=FALSE);
cat("\n\n# WN CDF of estimated FAPs:\n# FAP  CDF\n", file=logFile, append=TRUE);
write.table(data.frame(FAP=Xvalues,CDF=CDFvalues_WN), file=logFile, append=TRUE, row.names=FALSE, col.names=FALSE, quote=FALSE);


# write P-value pairs OUSS&WN to log file
cat(sprintf("# OUSS and WN analysis results for %d GPDD time series\n%s\n", nrow(OUSS_WN_RESULTS_TABLE), paste(colnames(OUSS_WN_RESULTS_TABLE),collapse="\t")), file=scatterFile, append=TRUE);
write.table(OUSS_WN_RESULTS_TABLE, file=scatterFile, append=TRUE, row.names=FALSE, col.names=FALSE, quote=FALSE, sep="\t");


# plot
plot(x=Xvalues, y=CDFvalues_OUSS, log="xy", type="l", ylab="CDF", xlab="OUSS eFAP", main=sprintf("CDF for OUSS FAP estimator\n(%d samples)",length(FAPs_OUSS)), cex=0.8, cex.main=0.9, xlim=range(Xvalues),ylim=range(Xvalues));
abline(a=0, b=1, col="grey");
plot(x=Xvalues, y=CDFvalues_WN, log="xy", type="l", ylab="CDF", xlab="WN eFAP", main=sprintf("CDF for WN FAP estimator\n(%d samples)",length(FAPs_WN)), cex=0.8, cex.main=0.9, xlim=range(Xvalues),ylim=range(Xvalues));
abline(a=0, b=1, col="grey");
dev.copy2pdf(file=paste0(plotDir, "/CDF_plot.pdf"));





